module Ubc.Parse.Syntax.Expression
( Expression(..)
, expressionParser
)
where

import Control.Monad ( (<$!>) )
import Data.Functor ( (<&>), ($>) )
import Data.Ratio ((%))

import Text.Parsec.Expr (Operator(Infix, Prefix), Assoc (AssocLeft, AssocRight), buildExpressionParser)
import Text.Parsec (ParsecT, (<|>), (<?>), choice, option, many1, digit, lookAhead, oneOf, char, notFollowedBy)

import Ubc.Parse.Syntax.Statement (blockExpression, Statement)
import Ubc.Parse.Syntax.Operators (BinaryOperator(..), UnaryOperator (..))

import qualified Ubc.Parse.Syntax.Language as UbcLanguage

data Expression = Binary BinaryOperator Expression Expression
        | Unary UnaryOperator Expression
        | ConstantInteger Integer
        | ConstantFraction Rational
        | FunctionCall String [Expression]
        | Variable String
        | If Expression Expression (Maybe Expression) -- if then else
        | Loop Expression Expression -- condition body
        | Block [Statement]
        deriving (Show)

operatorTable :: Monad m => [[Operator String u m Expression]]
operatorTable =
        [
                [ Infix (UbcLanguage.reservedOperator "*" $> Binary Multiply) AssocLeft
                , Infix (UbcLanguage.reservedOperator "/" $> Binary Divide) AssocLeft
                , Infix (UbcLanguage.reservedOperator "%" $> Binary Modulo) AssocLeft
                ]
        ,
                [ Infix (UbcLanguage.reservedOperator "+" $> Binary Plus) AssocLeft
                , Infix (UbcLanguage.reservedOperator "-" $> Binary Minus) AssocLeft
                ]
        ,
                [ Prefix (UbcLanguage.reservedOperator "!" $> Unary LogicNot)
                ]
        ,
                [ Infix (UbcLanguage.reservedOperator "<<" $> Binary ShiftLeft) AssocLeft
                , Infix (UbcLanguage.reservedOperator ">>" $> Binary ShiftRight) AssocLeft
                ]
        ,
                [ Infix (UbcLanguage.reservedOperator ">=" $> Binary GreaterEqual) AssocLeft
                , Infix (UbcLanguage.reservedOperator "<=" $> Binary LessEqual) AssocLeft
                , Infix (UbcLanguage.reservedOperator "<" $> Binary LessThan) AssocLeft
                , Infix (UbcLanguage.reservedOperator ">" $> Binary GreaterThan) AssocLeft
                ]
        ,
                [ Infix (UbcLanguage.reservedOperator "==" $> Binary Equal) AssocLeft
                , Infix (UbcLanguage.reservedOperator "!=" $> Binary NotEqual) AssocLeft
                ]
        ,       [ Infix (UbcLanguage.reservedOperator "&" $> Binary BitAnd) AssocLeft ]
        ,       [ Infix (UbcLanguage.reservedOperator "^" $> Binary BitXor) AssocLeft ]
        ,       [ Infix (UbcLanguage.reservedOperator "|" $> Binary BitOr) AssocLeft ]
        ,       [ Infix (UbcLanguage.reservedOperator "&&" $> Binary LogicAnd) AssocLeft ]
        ,       [ Infix (UbcLanguage.reservedOperator "||" $> Binary LogicOr) AssocLeft ]
        ,       [ Infix (UbcLanguage.reservedOperator "=" $> Binary Assign) AssocRight ]
        ]

expressionParser :: (Monad m) => ParsecT String u m Expression
expressionParser = buildExpressionParser operatorTable termParser <?> "expression"

numberTerm :: Monad m => ParsecT String u m Expression
numberTerm = do
        notFollowedBy $ char '0'
        integerDigits <- many1 digit
        choice
                [ decimalTerm integerDigits
                , return $ ConstantInteger (read integerDigits)
                ]
                <* UbcLanguage.whiteSpace

decimalTerm :: Monad m => [Char] -> ParsecT String u m Expression
decimalTerm integerDigits = do
        _ <- lookAhead $ oneOf "e."
        fractionalDigits <- option "" $ char '.' *> many1 digit
        powerSuffix <- option 1 $ char 'e' *> UbcLanguage.integer
        let numeratorPower = if powerSuffix > 0 then powerSuffix else 1
        let denominatorPower = if powerSuffix < 0 then powerSuffix else 1
        let numeratorInteger = read $ integerDigits ++ fractionalDigits
        let denominatorInteger = 10 ^ length fractionalDigits

        return $ ConstantFraction $ numeratorInteger ^ numeratorPower % denominatorInteger ^ denominatorPower


termParser :: Monad m => ParsecT String u m Expression
termParser = UbcLanguage.parens expressionParser
        <|> numberTerm
        <|> fmap ConstantInteger UbcLanguage.integer
        <|> conditionalExpression "if" id
        <|> conditionalExpression "unless" (Unary LogicNot)
        <|> loopExpression "while" id
        <|> loopExpression "until" (Unary LogicNot)
        <|> blockExpression
        <|> functionCallOrVariable
        <?> "term"

loopExpression :: Monad m => String -> (Expression -> Expression) -> ParsecT String u m Expression
loopExpression startKeyword conditionWrapper = do
        _ <- UbcLanguage.reserved startKeyword
        condition <- conditionWrapper <$!> expressionParser
        Loop condition <$> expressionParser

conditionalExpression :: Monad m => String -> (Expression -> Expression) -> ParsecT String u m Expression
conditionalExpression startKeyword conditionWrapper = do
        _ <- UbcLanguage.reserved startKeyword
        condition <- conditionWrapper <$!> expressionParser
        _ <- UbcLanguage.reserved "then"
        then_ <- expressionParser
        else_ <- option Nothing (UbcLanguage.reserved "else" >> expressionParser <&> Just)

        return $ If condition then_ else_

functionCallOrVariable :: Monad m => ParsecT String u m Expression
functionCallOrVariable = do
        name <- UbcLanguage.identifier
        choice
                [ FunctionCall name <$!> UbcLanguage.parens (UbcLanguage.commaSeparated expressionParser)
                , return $ Variable name
                ]

